"""
This file provides visualization utilities for ome-zarr file, similar to napari-ome-zarr,
but includes features like displaying zip ome zarr files
"""
from typing import Callable

import napari
import zarr
import dask.array as da
import cvpl_tools.ome_zarr.io as ome_io
import copy


# -------------Part 1: convenience functions, for adding ome zarr images using paths--------------


def preprocess_path_and_array_slices(path: str, kwargs: dict | None):
    """Preprocess array slices from path and update kwargs depending on what is found"""
    path, slices = ome_io.split_query_string(path)
    if kwargs is None:
        kwargs = dict()
    else:
        kwargs = copy.copy(kwargs)
        if slices is not None:
            to_append = kwargs.pop('array_slices', tuple())
            kwargs['array_slices'] = slices + to_append
    return path, kwargs


def add_ome_zarr_group_from_path(viewer: napari.Viewer, path: str, use_zip: bool | None = None,
                                 merge_channels=False, kwargs=None, lbl_kwargs=None):
    """Add an ome zarr group to napari viewer from given group path.

    This is a combination of load_zarr_group_from_path() and add_ome_zarr_group() functions.
    """
    assert isinstance(merge_channels, bool)
    path, kwargs = preprocess_path_and_array_slices(path, kwargs)

    zarr_group = ome_io.load_zarr_group_from_path(path, 'r', use_zip)
    add_ome_zarr_group(viewer,
                       zarr_group,
                       merge_channels=merge_channels,
                       kwargs=kwargs,
                       lbl_kwargs=lbl_kwargs)


def add_ome_zarr_array_from_path(viewer: napari.Viewer, path: str, use_zip: bool | None = None,
                                 merge_channels=False, kwargs=None):
    """Add an ome zarr array to napari viewer from given array path.

    This is a combination of load_zarr_array_from_path() and add_ome_zarr_group() functions.
    """
    assert isinstance(merge_channels, bool)
    path, kwargs = preprocess_path_and_array_slices(path, kwargs)

    zarr_group = ome_io.load_zarr_group_from_path(path, 'r', use_zip)
    add_ome_zarr_array(viewer, zarr_group, merge_channels=merge_channels, **kwargs)


# ------------------------Part 2:adding ome zarr files using zarr group---------------------------


def add_ome_zarr_group(viewer: napari.Viewer,
                       zarr_group: zarr.hierarchy.Group,
                       merge_channels=False,
                       kwargs: dict = None,
                       lbl_kwargs: dict = None):
    """Add an ome zarr image (if exists) along with its labels (if exist) to viewer.

    Args:
        viewer: Napari viewer object to attach image to
        zarr_group: The zarr group that contains the ome zarr file
        merge_channels: If True, display the image as one layers instead of a layer per channel
        kwargs: dictionary, keyword arguments to be passed to viewer.add_image for root image
        lbl_kwargs: dictionary, keyword arguments to be passed to viewer.add_image for label images
    """
    assert isinstance(merge_channels, bool)
    if kwargs is None:
        kwargs = {}
    if '0' in zarr_group:
        add_ome_zarr_array(viewer,
                           zarr_group,
                           merge_channels=merge_channels,
                           **kwargs)
    if 'labels' in zarr_group:
        if lbl_kwargs is None:
            lbl_kwargs = {}
        lbls_group = zarr_group['labels']
        for group_key in lbls_group.group_keys():
            lbl_group = lbls_group[group_key]
            add_ome_zarr_array(viewer, lbl_group, name=group_key, **lbl_kwargs)


def _add_ch(viewer: napari.Viewer,
            zarr_group: zarr.hierarchy.Group,
            arr_from_group: Callable[[zarr.hierarchy.Group], da.Array],
            start_level: int = 0,
            is_label=False, **kwargs):
    """Adds a channel or a set of channels to viewer"""
    multiscale = []
    while True:
        i = len(multiscale) + start_level
        i_str = str(i)
        if i_str in zarr_group:  # by ome zarr standard, image pyramid starts from 0 to NLEVEL - 1
            multiscale.append(arr_from_group(zarr_group[i_str]))
        else:
            break

    multiscale_flag = len(multiscale) > 1
    if not multiscale_flag:
        # newer version of napari seem to treat this as a single image with a width 1 axis in front, so unpack it
        multiscale = multiscale[0]
    if is_label:
        viewer.add_labels(multiscale, multiscale=multiscale_flag, **kwargs)
    else:
        viewer.add_image(multiscale, multiscale=multiscale_flag, **kwargs)


def add_ome_zarr_array(viewer: napari.Viewer,
                       zarr_group: zarr.hierarchy.Group,
                       merge_channels=False,
                       start_level: int = 0,
                       is_label=False,
                       **kwargs):
    """Add a multiscale ome zarr image or label to viewer.

    The first channel is assumed to be the channel dimension. This is relevant only if merge_channels=False

    Args:
        viewer (napari.Viewer): Napari viewer object to attach image to.
        zarr_group (zarr.hierarchy.Group): The zarr group that contains the ome zarr file.
        merge_channels: If True, display the image as one layers instead of a layer per channel
        start_level (int): The lowest level (highest resolution) to be added, default to 0
        is_label (bool): If True, display the image as label; this is suitable for instance segmentation
            masks where the results need a distinct color for each number
        ``**kwargs``: Keyword arguments to be passed to viewer.add_image for root image.
    """
    assert isinstance(merge_channels, bool)
    arr_shape = da.from_zarr(zarr_group[str(start_level)]).shape
    ndim = len(arr_shape)

    if kwargs.get('array_slices') is not None:
        slices = kwargs['array_slices']

        def arr_from_group(g):
            return da.from_array(g)[slices]
    else:
        def arr_from_group(g):
            return da.from_array(g)

    if ndim <= 2 or merge_channels:
        _add_ch(viewer, zarr_group, arr_from_group, start_level, is_label, **kwargs)
    else:
        nchan = arr_shape[0]
        assert nchan <= 20, (f'More than 20 channels (nchan={nchan}) found for add_ome_zarr_array, are you sure '
                             f'this is not a mistake? The function takes in a merge_channels option for images '
                             f'that are not multi-channel; by default the first axis of the image will be '
                             f'treated as channel dimension!')
        name = kwargs.get('name', 'ome_zarr')
        for i in range(nchan):
            ch_name = f'{name}_ch{i}'
            kwargs['name'] = ch_name
            _add_ch(viewer, zarr_group, lambda g: da.take(arr_from_group(g), indices=i, axis=0),
                    start_level, is_label, **kwargs)



from typing import Sequence

from cvpl_tools.im.fs import CachePointer
from cvpl_tools.im.seg_process import SegProcess, lc_interpretable_napari
import numpy as np
import numpy.typing as npt
from cvpl_tools.im.ndblock import NDBlock
import cvpl_tools.im.algorithms as algorithms
import dask.array as da
from scipy.ndimage import find_objects
import dask
from dask.distributed import print as dprint


class DirectOSToLC(SegProcess):
    """Convert a 0-N contour mask to a list of N centroids, one for each contour

    The converted list of centroids is in the same order as the original contour order (The contour labeled
    1 will come first and before contour labeled 2, 3 and so on)
    """

    def __init__(self,
                 min_size: int = 0,
                 reduce: bool = False,
                 is_global: bool = False,
                 ex_statistics: Sequence[str] = tuple()):
        super().__init__()
        self.reduce = reduce
        self.min_size = min_size
        self.is_global = is_global
        self.ex_statistics = ex_statistics
        self.full_statistics_map = dict(
            nvoxel=0,
            edge_contact=1,
            id=2
        )  # these columns are appended to feature array; self.ex_statistics selects from these to return
        self.global_full_statistics_map = dict(
            nvoxel=0,
            id=1
        )
        self.ret_cols = tuple(self.full_statistics_map[name] for name in self.ex_statistics)

        self._ndblock: NDBlock = None
        self._reduced_features = None
        self._reduced_np_features = None

    @property
    def reduced_features(self):
        if self._reduced_features is None:
            self._reduced_features = self._ndblock.reduce(force_numpy=False)
        return self._reduced_features

    @property
    def reduced_np_features(self):
        if self._reduced_np_features is None:
            rf = self.reduced_features
            if isinstance(rf, da.Array):
                rf = rf.compute()
            self._reduced_np_features = rf
        return self._reduced_np_features

    def np_features(self, block: npt.NDArray[np.int32], block_info=None) \
            -> npt.NDArray[np.float64]:
        if block_info is not None:
            mat_width = block.ndim + len(self.full_statistics_map)
            idx_max = np.array(tuple(d - 1 for d in block.shape), dtype=np.int64)

            slices = block_info[0]['array-location']
            _contours_argwhere, _ids = algorithms.npindices_from_os(block, is_sparse=True)
            contours_argwhere, ids = [], []
            for i in range(len(_contours_argwhere)):
                contour = _contours_argwhere[i]
                assert contour.shape[1] == block.ndim, (f'Contour shape {contour.shape} returned by npindices_from_os '
                                                        f'does not match block.ndim={block.ndim}')
                if len(contour) > self.min_size:
                    contours_argwhere.append(contour)
                    ids.append(_ids[i])
            ids = np.array(ids, dtype=_ids.dtype)
            lc = [contour.astype(np.float64).mean(axis=0) for contour in contours_argwhere]

            is_empty = len(lc) == 0
            if is_empty:
                lc = np.zeros((0, mat_width), dtype=np.float64)
            else:
                tmp = lc
                lc = np.zeros((len(lc), mat_width), dtype=np.float64)
                lc[:, :block.ndim] = tmp
            if slices is not None and not is_empty:
                start_pos = np.array([slices[i].start for i in range(len(slices))], dtype=np.float64)
                assert not np.isnan(start_pos).any(), 'nan should not be present in slice() objects for this!'
                lc[:, :block.ndim] += start_pos[None, :]

            # append extra statistics columns
            for name, j in self.full_statistics_map.items():
                if name == 'nvoxel':
                    col = [contour.shape[0] for contour in contours_argwhere]
                elif name == 'edge_contact':
                    col = []
                    for contour in contours_argwhere:
                        on_edge = (contour == 0).astype(np.uint8) + (contour == idx_max[None, :]).astype(np.uint8)
                        col.append(on_edge.sum().item() > 0)
                elif name == 'id':
                    col = ids
                else:
                    raise ValueError(f'Unrecognized name at index {j}: {name}')
                lc[:, block.ndim + j] = col

            return lc
        else:
            return np.zeros(block.shape, dtype=np.float64)

    def feature_forward(self, im: npt.NDArray[np.int32] | da.Array) -> NDBlock[np.float64]:
        return NDBlock.map_ndblocks([NDBlock(im)], self.np_features, out_dtype=np.float64)

    def aggregate_by_id(self):
        """Aggregate self._ndblock by id"""

        ref_ndblock: NDBlock = self._ndblock
        ndim = ref_ndblock.get_ndim()
        block_indices = ref_ndblock.get_block_indices()
        slices_list = ref_ndblock.get_slices_list()
        chunk_shape = np.array(tuple(s.stop - s.start for s in slices_list[0]), dtype=np.float64)
        recons = {ind: [] for ind in block_indices}

        rf = self.reduced_np_features
        ordered = rf[np.argsort(rf[:, -1])]
        if rf.shape[0] == 0:
            cnt_ranges = []
        else:
            cnt_ranges = list(find_objects(ordered[:, -1].astype(np.int32)))

        nvoxel_ind = self.full_statistics_map['nvoxel']
        for i, rg in enumerate(cnt_ranges):
            if rg is None:
                continue
            # lbl = i + 1
            subrf = rf[rg]
            nvoxel = subrf[:, ndim + nvoxel_ind]
            nvoxel_tot = nvoxel.sum()
            centroid = (subrf[:, :ndim] * nvoxel[:, None]).sum(axis=0) / nvoxel_tot

            row = centroid.tolist()
            row.append(nvoxel_tot)
            row.append(i)
            row = np.array(row, dtype=np.float64)
            ind = tuple(np.floor(centroid / chunk_shape).astype(np.int32).tolist())
            recons[ind].append(row)

        for i in range(len(slices_list)):
            ind = block_indices[i]
            rows = recons[ind]
            if len(rows) == 0:
                rows = np.zeros((0, ndim + len(self.global_full_statistics_map)), dtype=np.float64)
            else:
                rows = np.array(recons[ind], dtype=np.float64)
            recons[ind] = (rows, slices_list[i])
        self._ndblock = NDBlock.create_from_dict_and_properties(recons,
                                                                ref_ndblock.get_properties() | dict(is_numpy=True))

    def forward(self,
                im: npt.NDArray[np.int32] | da.Array,
                cptr: CachePointer,
                viewer_args: dict = None) -> NDBlock[np.float64] | npt.NDArray[np.float64]:
        if viewer_args is None:
            viewer_args = {}
        viewer = viewer_args.get('viewer', None)

        cdir = cptr.subdir()

        def compute():
            self._ndblock = cdir.cache_im(fn=lambda: self.feature_forward(im),
                                          cache_level=1,
                                          cid='block_level_lc_ndblock')
            if self.is_global:
                # update and aggregate the rows in ndblock that correspond to the same contour
                self.aggregate_by_id()
                self._reduced_features = None
                self._reduced_np_features = None
            return self._ndblock

        self._ndblock = cdir.cache_im(fn=compute,
                                      cache_level=1,
                                      cid='lc_ndblock')

        if viewer and viewer_args.get('display_points', True):
            if self.is_global:
                extras = list(self.global_full_statistics_map.keys())
            else:
                extras = list(self.full_statistics_map.keys())
            lc_interpretable_napari(
                'os_to_lc_centroids',
                self.reduced_np_features,
                viewer,
                im.ndim,
                extras
            )

        ret_cols = tuple(range(im.ndim)) + tuple(r + im.ndim for r in self.ret_cols)
        if self.reduce:
            ret = self.reduced_features[:, ret_cols]
        else:
            ret = self._ndblock.select_columns(ret_cols)
        self._ndblock = None
        self._reduced_features = None
        self._reduced_np_features = None

        return ret



"""
Add support for writing dask array to ome zarr zip file, and avoid repeated computation as
discussed in https://github.com/ome/ome-zarr-py/issues/392
"""
from typing import Sequence

import ome_zarr
import ome_zarr.io
import cvpl_tools.ome_zarr.ome_zarr_writer_patched as writer
import os
import shutil
import zarr
import dask.array as da
import numpy as np
import urllib.parse
import numcodecs


# ----------------------------------Part 1: utilities---------------------------------------


def encode_path_as_filename(path):
    encoded_path = urllib.parse.quote(path, safe='')
    return encoded_path


def load_zarr_group_from_path(path: str,
                              mode: str | None = None,
                              use_zip: bool | None = None,
                              level: int | None = None) -> zarr.Group:
    """Loads either a zarr folder or zarr zip file into a zarr group.

    Args:
        path: path to the zarr folder or zip to be opened
        mode: file open mode e.g. 'r', only pass this if the file is a zip file
        use_zip: if True, treat path as a zip; if False, treat path as a folder; if None, use path to determine
            file type
        level: If None (default), load the entire ome zarr; if an int is provided, load the corresponding level
            in the ome zarr array
    Returns:
        the opened zarr group
    """
    if use_zip is None:
        use_zip = path.endswith('.zip')
    if use_zip:
        store = zarr.ZipStore(path, mode=mode)
    else:
        store = zarr.DirectoryStore(path)
    zarr_group = zarr.open(store, mode=mode)
    if level is not None:
        assert isinstance(level, int)
        zarr_group = zarr_group[str(level)]

    return zarr_group


def split_query_string(path: str) -> tuple[str, tuple[slice, ...] | None]:
    """Split a url to path and an optional query dict

    Examples:
        split_query_string('path/to/file') -> ('path/to/file', None)
        split_query_string('path/to/file?slices=[3, :2]') -> ('path/to/file', tuple(3, slice(None, 2)))

    Args:
        path: url to be splitted

    Returns:
        splitted path and query string; if path does not contain a query string, then the second returned arg is None.
    """
    if '?' in path:
        # parse the query string at the end
        path, query = path.split('?')
        query = urllib.parse.parse_qs(query)

        slices = query['slices'][0]
        assert slices[0] == '[' and slices[-1] == ']', f'Slices must start with [ and end with ], got slices={slices}'
        slices = slices[1:-1].split(',')
        result = []
        for comp in slices:
            if ':' in comp:
                result.append(slice(*[int(x) if x else None for x in comp.split(':')]))
            else:
                result.append(int(comp))
        slices = tuple(result)
    else:
        slices = None
    return path, slices


def filename_from(path: str) -> str:
    """Strip off query in a url that may contain query e.g. 'path/to/file?slices=[1]'

    Args:
        path: url to be splitted

    Returns:
        path with trailing query stripped
    """
    return split_query_string(path)[0]


def load_dask_array_from_path(path: str,
                              mode: str | None = None,
                              use_zip: bool | None = None,
                              level: int | None = None) -> da.Array:
    """Loads either a zarr folder or zarr zip file into a dask array.

    Compared to load_zarr_group_from_path, this function allows specifying which slice and channel to read using
    a query string syntax (idea thanks to Davis Bennett in the thread
    https://forum.image.sc/t/loading-only-one-channel-from-an-ome-zarr/97798)

    Example:
        Loading an ome zarr array of shape (2, 200, 1000, 1000) using different slices::

            arr_original = load_dask_array_from_path('file.ome.zarr', level=0)  # shape=(2, 200, 1000, 1000)
            arr1 = load_dask_array_from_path('file.ome.zarr?slices=[0]', level=0)  # shape=(200, 1000, 1000)
            arr2 = load_dask_array_from_path('file.ome.zarr?slices=[:, :100]', level=0)  # shape=(2, 100, 1000, 1000)
            arr3 = load_dask_array_from_path('file.ome.zarr?slices=[0:1, 0, -1:, ::2]', level=0)  # shape=(1, 1, 500)

        Essentially, Python multi-index slicing can be done and the effect is similar to torch or numpy indexing using
        slices.

    Args:
        path: path to the zarr folder or zip to be opened
        mode: file open mode e.g. 'r', only pass this if the file is a zip file
        use_zip: if True, treat path as a zip; if False, treat path as a folder; if None, use path to determine
            file type
        level: If None (default), load the entire ome zarr; if an int is provided, load the corresponding level
            in the ome zarr array
    Returns:
        the opened zarr group
    """
    path, slices = split_query_string(path)

    zarr_group = load_zarr_group_from_path(path, mode, use_zip, level)
    arr = da.from_array(zarr_group)

    if slices is not None:
        arr = arr[slices]

    return arr


# --------------------------------Part 2: write image-------------------------------------


def _get_coord_transform_yx_for_write(ndim, MAX_LAYER) -> list:  # down sampling by 2 each layer
    assert ndim >= 2, 'Can not 2d down sample an image of dimension less than 2d!'
    coordinate_transformations = []
    for layer in range(MAX_LAYER + 1):
        coordinate_transformations.append(
            [{'scale': [1.] * (ndim - 2) + [(2 ** layer) * 1., (2 ** layer) * 1.],  # image-pyramids in XY only
              'type': 'scale'}])
    return coordinate_transformations


def _get_axes_for_write(ndim: int) -> list:
    axes = ['x', 'y', 'z', 'c']  # usually ['x', 'y', 'z', 'c', 't'] but we don't need that for our use
    return list(reversed(axes[:ndim]))


def write_ome_zarr_image_direct(zarr_group: zarr.Group,
                                da_arr: da.Array | None = None,
                                lbl_arr: da.Array | None = None,
                                lbl_name: str | None = None,
                                MAX_LAYER: int = 3,
                                storage_options: dict = None,
                                lbl_storage_options: dict = None):
    """Direct write of dask array to target ome zarr group (can not be a zip)

    Args:
        zarr_group: The output zarr group which we will write an ome zarr image to
        da_arr: (dask array) The content of the image to write
        lbl_arr: If provided, this is the array to write at zarr_group['labels'][lbl_name]
        lbl_name: name of the label array subgroup
        MAX_LAYER: The maximum layer of down sampling; starting at layer=0
        storage_options: options for storing the image
        lbl_storage_options: options for storing the labels
    """
    if storage_options is None:
        compressor = numcodecs.Blosc(cname='lz4', clevel=9, shuffle=numcodecs.Blosc.BITSHUFFLE)
        storage_options = dict(
            dimension_separator='/',
            compressor=compressor
        )
    if da_arr is not None:
        # assert the group is empty, since we are writing a new group
        for mem in zarr_group:
            raise ValueError('ZARR group is Non-empty, please remove the original ZARR before running the program to '
                             f'create synthetic data. ZARR group: {zarr_group}')

    scaler = ome_zarr.scale.Scaler(max_layer=MAX_LAYER, method='nearest')
    if da_arr is not None:
        writer.write_image(image=da_arr,
                           group=zarr_group,
                           scaler=scaler,
                           coordinate_transformations=_get_coord_transform_yx_for_write(da_arr.ndim, MAX_LAYER),
                           storage_options=storage_options,
                           axes=_get_axes_for_write(da_arr.ndim))

    if lbl_arr is not None:
        assert lbl_name is not None, ('ERROR: Please provide lbl_name along when writing labels '
                                      '(lbl_arr is not None)')

        # we could just use ['c', 'z', 'y', 'x'], however, napari ome zarr can't handle channel types but only space
        # type axes. So we need to fall back to manual definition, avoid 'c' which defaults to a channel type
        lbl_axes = [{'name': ch, 'type': 'space'} for ch in _get_axes_for_write(lbl_arr.ndim)]

        if lbl_storage_options is None:
            compressor = numcodecs.Blosc(cname='lz4', clevel=9, shuffle=numcodecs.Blosc.BITSHUFFLE)
            lbl_storage_options = dict(compressor=compressor)
        writer.write_labels(labels=lbl_arr,
                            group=zarr_group,
                            scaler=scaler,
                            name=lbl_name,
                            coordinate_transformations=_get_coord_transform_yx_for_write(lbl_arr.ndim, MAX_LAYER),
                            storage_options=lbl_storage_options,
                            axes=lbl_axes)
        # ome_zarr.writer.write_label_metadata(group=g,
        #                      name=f'/labels/{lbl_name}',
        #                      properties=properties)


def write_ome_zarr_image(ome_zarr_path: str,
                         tmp_path: str = None,  # provide this if making a zip file
                         da_arr: da.Array | None = None,
                         lbl_arr: da.Array | None = None,
                         lbl_name: str | None = None,
                         make_zip: bool | None = None,
                         MAX_LAYER: int = 0,
                         logging=False,
                         storage_options: dict = None,
                         lbl_storage_options: dict = None):
    """Write dask array as an ome zarr

    For writing to zip file: due to dask does not directly support write to zip file, we instead create a temp ome zarr
    output and copy it into a zip after done. This is why tmp_path is required if make_zip=True

    Args:
        ome_zarr_path: The path to target ome zarr folder, or ome zarr zip folder if make_zip is True
        tmp_path: temporary files will be stored under this,
        da_arr: If provided, this is the array to write at {ome_zarr_path}
        lbl_arr: If provided, this is the array to write at {ome_zarr_path}/labels/{lbl_name}
        lbl_name: name of the folder of the label array
        make_zip: bool, if True the output is a zip; if False a folder; if None, then determine based on file suffix
        MAX_LAYER: The maximum layer of down sampling; starting at layer=0
        logging: If true, print message when job starts and ends
        storage_options: options for storing the image
        lbl_storage_options: options for storing the labels
    """
    if tmp_path is not None:
        os.makedirs(tmp_path, exist_ok=True)

    if make_zip is None:
        make_zip = ome_zarr_path.endswith('.zip')
    if logging:
        print('Writing Folder.')
    if make_zip:
        encode_name = encode_path_as_filename(ome_zarr_path)
        folder_ome_zarr_path = f'{tmp_path}/{encode_name}.temp'
        if os.path.exists(folder_ome_zarr_path):
            shutil.rmtree(folder_ome_zarr_path)
    else:
        folder_ome_zarr_path = ome_zarr_path

    store = ome_zarr.io.parse_url(folder_ome_zarr_path, mode='w').store
    g = zarr.group(store)
    write_ome_zarr_image_direct(g, da_arr, lbl_arr, lbl_name, MAX_LAYER=MAX_LAYER,
                                storage_options=storage_options,
                                lbl_storage_options=lbl_storage_options)
    if logging:
        print('Folder is written.')
    store.close()

    if make_zip:
        if logging:
            print('Writing zip.')
        store = ome_zarr.io.parse_url(folder_ome_zarr_path,
                                      mode='r').store  # same folder but this time we open it in read mode
        g = zarr.group(store)
        with zarr.ZipStore(ome_zarr_path, mode='w') as target_store:
            target_g = zarr.group(target_store)
            zarr.copy_all(g, target_g)

        if logging:
            print('Zip is written.')
        # remove the temp folder
        shutil.rmtree(folder_ome_zarr_path)


# ------------------------------Part 3: synthetic dataset-----------------------------------


def generate_synthetic_dataset(ome_zarr_path: str,
                               tmp_path: str = None,
                               arr_sz: tuple = (2, 224, 1600, 2048),
                               chunks: tuple = (1, 1, 1024, 1024),
                               make_zip: bool | None = None,
                               MAX_LAYER=0) -> zarr.Group:
    """Generate a 4d synthetic test ome zarr image physically stored in ome_zarr_path.

    Args:
        ome_zarr_path: Where to store the generated test image
        tmp_path: The temporary path for write if making a zip file image
        arr_sz: The size of the synthetic image
        chunks: The chunk size of the image
        make_zip: If True, make a physical zip storage of instead of a directory of ome zarr image
        MAX_LAYER: The maximum down sampling layer
    Returns:
        An opened zarr group of the newly generated synthetic image
    """
    arr: da.Array = da.zeros(arr_sz, dtype=np.uint16, chunks=chunks)

    def process_block(block, block_info=None):
        if block_info is not None:
            # calculate (global) indices array for each pixel
            block_slice = block_info[0]['array-location']
            indices = np.indices(block.shape)
            for dim in range(indices.shape[0]):
                indices[dim] += block_slice[dim][0]
        else:
            return np.zeros(block.shape, dtype=np.uint16)
        # now, create balls in the block
        sq = np.zeros(block.shape)  # distance squared
        for dim in range(1, indices.shape[0]):  # every dim except channel dim which does not have distance
            sq += np.power(indices[dim], 2.) * .0002
        for dim in range(1, indices.shape[0]):  # every dim except channel dim which does not have distance
            indices[dim] %= 32
            sq += np.power(indices[dim] - 15.5, 2.)
        im = np.array(np.clip(1200. - sq * 15., 0., 1200.), dtype=np.uint16)
        return im

    arr = arr.map_blocks(process_block, dtype=np.uint16)
    write_ome_zarr_image(ome_zarr_path, tmp_path, da_arr=arr, make_zip=make_zip, MAX_LAYER=MAX_LAYER)
    group = load_zarr_group_from_path(ome_zarr_path, mode='r', use_zip=make_zip)
    return group


def dask_checkerboard(chunks: Sequence[Sequence[int]]) -> da.Array:
    """Create a synthetic checkerboard pattern dask image

    Args:
        chunks: The chunks of the checkerboard, from which shape and result array chunk sizes will
            be calculated from

    Returns:
        A dask.Array of checkerboard patterns (0=black and 1=white) of type np.uint8; top-left is
            black
    """
    shape = tuple(sum(c) for c in chunks)
    checkerboard: da.Array = da.zeros(shape, chunks=chunks, dtype=np.uint8)

    def map_fn(block, block_info=None):
        block_index = block_info[0]['chunk-location']
        color = sum(block_index) % 2
        return np.ones_like(block) * color

    checkerboard = checkerboard.map_blocks(map_fn, meta=np.zeros(tuple(), dtype=np.uint8), dtype=np.uint8)
    return checkerboard




"""
This file defines the NDBlock class and related enum, helper functions related to it
"""
from __future__ import annotations

import os
import abc
import enum
import json
import pickle
import numcodecs
import threading
from typing import Callable, Sequence, Any, Generic, TypeVar
import copy

import partd
from dask.distributed import print as dprint
import numpy as np
import numpy.typing as npt
import dask.array as da
import dask
import functools
import operator
import cvpl_tools.ome_zarr.io as cvpl_ome_zarr_io
from cvpl_tools.im.partd_server import SQLitePartd, SqliteServer


class ReprFormat(enum.Enum):
    """ReprFormat specifies all possible NDBlock formats to use.

    NUMPY_ARRAY = 0
    DASK_ARRAY = 1
    DICT_BLOCK_INDEX_SLICES = 2  # block can be either numpy or dask
    """

    NUMPY_ARRAY = 0
    DASK_ARRAY = 1
    DICT_BLOCK_INDEX_SLICES = 2  # block can be either numpy or dask


ElementType = TypeVar('ElementType')


def dumps_numpy(arr: npt.NDArray, compressor) -> bytes:
    buf = pickle.dumps(arr)
    if compressor is not None:
        buf = compressor.encode(buf)
    return buf


def loads_numpy(buf: bytes, compressor) -> npt.NDArray:
    if compressor is not None:
        buf = compressor.decode(buf)
    return pickle.loads(buf)


class NDBlockLazyLoadDictBlockInfo:
    def __init__(self, db_path: str, compressor=None):
        """Initialize an info object containing information about how to load the NDBlock

        This class is for optimization purpose (lazy loading) where in some cases the computation is not
        necessary

        Args:
            db_path: path to the database object where NDBlock.save saves a DICT_BLOCK_INDEX_SLICES object
            compressor: A compressor that supports encode and decode interfaces on bytes to bytes compression
        """
        self.db_path = db_path
        self.compressor = compressor

    def get_db_path(self):
        return self.db_path

    def is_just_loaded(self):
        """Returns True if just loaded from files

        In some cases we save and then load an NDBlock to avoid double computation, but if the NDBlock
        is just loaded from disk then it is unnecessary to do save and load again. This function is for
        testing for that to avoid such situation
        """
        return True

    def materialize(self, ndblock: NDBlock, force_numpy=False):
        block_indices = ndblock.properties['block_indices']
        slices_list = ndblock.properties['slices_list']
        ndblock.arr = {}

        def load_block_with_compression(sqlite_file, block_id):
            return loads_numpy(sqlite_file.get(block_id, lock=False), self.compressor)

        if force_numpy or ndblock.properties['is_numpy']:
            sqlite_file = SQLitePartd(self.db_path)

            for i in range(len(slices_list)):
                block_index, slices = block_indices[i], slices_list[i]

                block_id = ('_'.join(str(i) for i in block_index)).encode('utf-8')
                block = load_block_with_compression(sqlite_file, block_id)

                ndblock.arr[block_index] = (block, slices)
            sqlite_file.close()
            ndblock.properties['is_numpy'] = True
        else:
            @dask.delayed
            def load_block(block_index):
                block_id = ('_'.join(str(i) for i in block_index)).encode('utf-8')
                sqlite_file = SQLitePartd(self.db_path)
                block = load_block_with_compression(sqlite_file, block_id)
                assert isinstance(block, np.ndarray)
                sqlite_file.close()
                return block

            for i in range(len(slices_list)):
                block_index, slices = block_indices[i], slices_list[i]

                ndblock.arr[block_index] = (load_block(block_index), slices)


class NDBlock(Generic[ElementType], abc.ABC):
    """This class represents an N-dimensional grid, each block in the grid is of arbitrary ndarray shape

    When the grid is of size 1 in all axes, it represents a Numpy array;
    When the grid has blocks of matching size on all axes, it represents a Dask array;
    In the general case the blocks can be of varying sizes e.g. A block of size (2, 2) may be neighboring to a block
    of size (5, 10)

    Currently, we assume block_index is a list always ordered in (0, ..., 0), (0, ..., 1)... (N, ..., M-1) Increasing
    from the side of the tail first
    """

    def __init__(self, arr: npt.NDArray | da.Array | NDBlock | None):
        """Initializes an NDBlock object

        Args:
            arr: If Numpy, create a NDBlock with repr_format=ReprFormat.NUMPY_ARRAY; if Dask, create a NDBlock
                with repr_format=ReprFormat.DASK_ARRAY; if arr is an NDBlock, then a copy of arr will be created
        """
        self.arr = None
        self.properties = None

        if arr is None:
            return  # don't fully initialize; used in case properties can't be inferred from arr

        if isinstance(arr, np.ndarray):
            self.arr = arr
            self._set_properties_by_numpy_array(arr)
        elif isinstance(arr, da.Array):
            self.arr = arr
            self._set_properties_by_dask_array(arr)
        else:
            assert isinstance(arr, NDBlock), f'Unexpected type {type(arr)}'
            self.arr = arr.arr
            self.properties = NDBlock._copy_properties(arr.properties)

    @staticmethod
    def create_from_dict_and_properties(d: dict, properties: dict[str, Any]):
        ndblock = NDBlock(None)
        ndblock.arr = d
        ndblock.properties = NDBlock._copy_properties(properties)
        is_numpy = False
        for block_index, (block, slices) in d.items():
            is_numpy = isinstance(block, np.ndarray)
            break
        assert properties['is_numpy'] == is_numpy, (f"Got is_numpy={properties['is_numpy']} in properties but "
                                                    f"is_numpy={is_numpy} for dictionary of blocks")
        return ndblock

    @staticmethod
    def properties_consistency_check(properties: dict[str, Any]):
        block_indices = properties['block_indices']
        assert isinstance(block_indices, list)
        assert isinstance(block_indices[0], tuple)

        slices_list = properties['slices_list']
        assert isinstance(slices_list, list)
        assert isinstance(slices_list[0], tuple)
        assert isinstance(slices_list[0][0], slice)

        repr_format = properties['repr_format']
        assert isinstance(repr_format, ReprFormat)

        numblocks = properties['numblocks']
        assert isinstance(numblocks, tuple)
        assert isinstance(numblocks[0], int)

    def get_arr(self):
        self.ensure_materialized()
        return self.arr

    def ensure_materialized(self, **kwargs):
        if isinstance(self.arr, NDBlockLazyLoadDictBlockInfo):
            self.arr.materialize(self, **kwargs)

    @staticmethod
    def save_properties(file: str, properties: dict):
        with open(file, mode='w') as outfile:
            # convert repr_format and dtype to appropriate format for serialization
            # reference: https://stackoverflow.com/questions/47641404/serializing-numpy-dtype-objects-human-readable
            properties = copy.copy(properties)
            properties['repr_format'] = properties['repr_format'].value
            properties['dtype'] = np.dtype(properties['dtype']).str
            properties['slices_list'] = [tuple((s.start, s.stop) for s in si) for si in properties['slices_list']]

            block_indices = properties['block_indices']
            assert isinstance(properties['block_indices'], list), (f'Expected block indices to be of type list, '
                                                                   f'got {block_indices}')
            json.dump(properties, fp=outfile, indent=2)

    @staticmethod
    def save(file: str, ndblock: NDBlock, storage_options: dict | None = None):
        """Save the NDBlock to the given path

        Will compute immediately if the ndblock is delayed dask computations

        Storage Opitons
            multiscale (int) = 0
                This only applies if ndblock is dask, will write multilevel if non-zero
            compressor = None
                The compressor to use to compress array or chunks

        Args:
            file: The file to save to
            ndblock: The block which will be saved
            storage_options: Specifies options in saving method and saved file format
        """
        if storage_options is None:
            storage_options = {}
        compressor = storage_options.get('compressor', None)

        ndblock.ensure_materialized()

        fmt = ndblock.get_repr_format()
        os.makedirs(file, exist_ok=False)
        if fmt == ReprFormat.NUMPY_ARRAY:
            if compressor is None:
                np.save(f'{file}/im.npy', ndblock.arr)
            else:
                with open(f'{file}/im_compressed.bin', 'wb') as binfile:
                    binfile.write(dumps_numpy(ndblock.arr, compressor=compressor))
        elif fmt == ReprFormat.DASK_ARRAY:
            MAX_LAYER = storage_options.get('multiscale') or 0
            cvpl_ome_zarr_io.write_ome_zarr_image(f'{file}/dask_im', da_arr=ndblock.arr,
                                                  make_zip=False, MAX_LAYER=MAX_LAYER,
                                                  storage_options=storage_options)
        else:
            server = SqliteServer(f'{file}/blocks_kvstore', nappend=len(ndblock.arr), available_memory=1e6,
                                  port_protocol=storage_options.get('port_protocol', 'tcp'))
            server_address = server.address

            @dask.delayed
            def save_block(block_index, block):
                block_id = ('_'.join(str(i) for i in block_index)).encode('utf-8')
                store = partd.Client(server_address)
                assert isinstance(block, np.ndarray), type(block)
                store.append({
                    block_id: dumps_numpy(block, compressor=compressor)
                })
                store.close()

            tasks = [save_block(block_index, block) for block_index, (block, _) in ndblock.arr.items()]
            dask.compute(*tasks)
            server.wait_join()
        NDBlock.save_properties(f'{file}/properties.json', ndblock.properties)

    @staticmethod
    def load_properties(file: str) -> dict:
        with open(file, mode='r') as infile:
            properties = json.load(fp=infile)

            properties['repr_format'] = ReprFormat(properties['repr_format'])
            properties['dtype'] = np.dtype(properties['dtype'])
            properties['block_indices'] = [tuple(idx) for idx in properties['block_indices']]
            properties['slices_list'] = [tuple(slice(s[0], s[1]) for s in si) for si in properties['slices_list']]
            properties['numblocks'] = tuple(properties['numblocks'])
        NDBlock.properties_consistency_check(properties)
        return properties

    @staticmethod
    def load(file: str, storage_options: dict | None = None) -> NDBlock:
        """Load the NDBlock from the given path.

        Storage Opitons
            compressor = None
                The compressor to use to compress array or chunks

        Args:
            file: The path to load from, same as used in the save() function
            storage_options: Specifies options in saving method and saved file format

        Returns:
            The loaded NDBlock. Guaranteed to have the same properties as the saved one, and the content of each
            block will be the same as when they are saved.
        """
        if storage_options is None:
            storage_options = {}
        compressor = storage_options.get('compressor', None)

        properties = NDBlock.load_properties(f'{file}/properties.json')

        fmt = properties['repr_format']
        ndblock = NDBlock(None)
        ndblock.properties = properties
        if fmt == ReprFormat.NUMPY_ARRAY:
            if compressor is None:
                ndblock.arr = np.load(f'{file}/im.npy')
            else:
                with open(f'{file}/im_compressed.bin', 'rb') as binfile:
                    ndblock.arr = loads_numpy(binfile.read(), compressor)
        elif fmt == ReprFormat.DASK_ARRAY:
            ndblock.arr = da.from_zarr(cvpl_ome_zarr_io.load_zarr_group_from_path(
                f'{file}/dask_im',
                mode='r',
                level=0
            ))
        else:
            ndblock.arr = NDBlockLazyLoadDictBlockInfo(db_path=f'{file}/blocks_kvstore', compressor=compressor)
        return ndblock

    def is_numpy(self) -> bool:
        """Returns True if this is Numpy array

        Note besides type ReprFormat.NUMPY, ReprFormat.DICT_BLOCK_INDEX_SLICES may have either Numpy arrays as each
        block, or Dask delayed objects each returning a Numpy array; in the former case is_numpy() will return True,
        and in the latter case is_numpy() will return False.

        Returns:
            Returns True if this is Numpy array
        """
        return self.properties['is_numpy']

    def get_properties(self) -> dict[str, Any]:
        return self.properties

    def get_repr_format(self) -> ReprFormat:
        return self.properties['repr_format']

    @property
    def ndim(self) -> int:
        return self.properties['ndim']

    def get_ndim(self) -> int:
        return self.properties['ndim']

    def get_numblocks(self) -> tuple[int]:
        return self.properties['numblocks']

    def get_block_indices(self) -> list:
        return self.properties['block_indices']

    def get_slices_list(self) -> list:
        return self.properties['slices_list']

    def get_dtype(self):
        return self.properties['dtype']

    def get_chunksize(self) -> tuple[int, ...]:
        """Get a single tuple of chunk size on each axis

        Returns:
            A tuple of int of length ndim
        """
        slices = self.properties['slices_list'][0]
        chunksize = tuple(s.stop - s.start for s in slices)
        return chunksize

    def as_numpy(self) -> npt.NDArray:
        other = NDBlock(self)
        other.to_numpy_array()
        return other.arr

    def as_dask_array(self, tmp_dirpath: str | None = None) -> da.Array:
        """Get a copy of the array value as dask array

        Args:
            tmp_dirpath: Avoid double computation, best provided if input format
                is ReprFormat.DICT_BLOCK_INDEX_SLICES and is dask instead of numpy
        Returns:
            converted/retrieved dask array
        """
        other = NDBlock(self)
        other.to_dask_array(tmp_dirpath)
        return other.arr

    @staticmethod
    def _copy_properties(properties: dict):
        NDBlock.properties_consistency_check(properties)
        return dict(
            repr_format=properties['repr_format'],
            ndim=properties['ndim'],
            numblocks=properties['numblocks'],
            block_indices=properties['block_indices'],
            slices_list=[item for item in properties['slices_list']],
            is_numpy=properties['is_numpy'],
            dtype=properties['dtype']
        )

    def _set_properties_by_numpy_array(self, arr: npt.NDArray):
        ndim: int = arr.ndim
        self.properties = dict(
            repr_format=ReprFormat.NUMPY_ARRAY,
            ndim=ndim,
            numblocks=(1,) * ndim,
            block_indices=[(0,) * ndim],
            slices_list=[tuple(slice(0, s) for s in arr.shape)],
            is_numpy=True,
            dtype=arr.dtype
        )
        NDBlock.properties_consistency_check(self.properties)

    def _set_properties_by_dask_array(self, arr: da.Array):
        ndim: int = arr.ndim
        numblocks = arr.numblocks
        block_indices = list(np.ndindex(*arr.numblocks))
        slices_list: list = da.core.slices_from_chunks(arr.chunks)

        self.properties = dict(
            repr_format=ReprFormat.DASK_ARRAY,
            ndim=ndim,
            numblocks=numblocks,
            block_indices=block_indices,
            slices_list=slices_list,
            is_numpy=False,
            dtype=arr.dtype
        )
        NDBlock.properties_consistency_check(self.properties)

    # ReprFormat conversion functions

    def to_numpy_array(self):
        """Convert representation format to numpy array"""
        rformat = self.properties['repr_format']
        if rformat == ReprFormat.NUMPY_ARRAY:
            return

        self.ensure_materialized()

        if rformat == ReprFormat.DICT_BLOCK_INDEX_SLICES:
            # TODO: optimize this
            self.to_dask_array()
            rformat = self.get_repr_format()

        assert rformat == ReprFormat.DASK_ARRAY
        self.arr = self.arr.compute()
        self._set_properties_by_numpy_array(self.arr)  # here some blocks may be merged to one

        self.properties['repr_format'] = ReprFormat.NUMPY_ARRAY
        self.properties['is_numpy'] = True

    def to_dask_array(self, tmp_dirpath: str | None = None):
        """Convert representation format to dask array

        Args:
            tmp_dirpath: An empty path to create a temporary save file, best provided if input format
                is ReprFormat.DICT_BLOCK_INDEX_SLICES and is dask instead of numpy, and you want to
                avoid repeated computations
        """

        rformat = self.properties['repr_format']
        if rformat == ReprFormat.DASK_ARRAY:
            return

        just_loaded_flag = isinstance(self.arr, NDBlockLazyLoadDictBlockInfo) and self.arr.is_just_loaded()

        if rformat == ReprFormat.NUMPY_ARRAY:
            self.ensure_materialized()
            self.arr = da.from_array(self.arr)
        else:
            assert rformat == ReprFormat.DICT_BLOCK_INDEX_SLICES
            numblocks = self.get_numblocks()
            dtype = self.get_dtype()

            if not self.properties['is_numpy'] and tmp_dirpath is not None and not just_loaded_flag:
                """Explanation of the branching statement: this if branch is optional and purely for 
                optimization purpose i.e. some dask computation may be triggered twice unexpectedly and 
                this avoids it
                
                not self.properties['is_numpy']
                - If numpy, the array is already computed and no double computation possible
                
                tmp_dirpath is not None
                - If tmp_dirpath is None, then we don't have a place to put our temporary save files
                
                not just_loaded_flag
                - Tests if the array if just loaded from disk; if it is, then no double computation possible since 
                    there are no intermediate steps in the first place
                """
                if not os.path.exists(tmp_dirpath):
                    NDBlock.save(tmp_dirpath, self)
                ndblock_to_be_combined = NDBlock.load(tmp_dirpath)
            else:
                ndblock_to_be_combined = self
            ndblock_to_be_combined.ensure_materialized()

            # reference: https://github.com/dask/dask-image/blob/adcb217de766dd6fef99895ed1a33bf78a97d14b/dask_image/ndmeasure/__init__.py#L299
            ndlists = np.empty(numblocks, dtype=object)
            for block_index, (block, slices) in ndblock_to_be_combined.arr.items():
                block_shape = tuple(s.stop - s.start for s in slices)
                if not isinstance(block, np.ndarray):
                    block = da.from_delayed(block, shape=block_shape, dtype=dtype)
                ndlists[block_index] = block
            ndlists = ndlists.tolist()
            self.arr = da.block(ndlists)
        self.properties['repr_format'] = ReprFormat.DASK_ARRAY
        self.properties['is_numpy'] = False

    def to_dict_block_index_slices(self):
        """Convert representation format to ReprFormat.DICT_BLOCK_INDEX_SLICES"""
        rformat = self.properties['repr_format']
        if rformat == ReprFormat.DICT_BLOCK_INDEX_SLICES:
            return

        self.ensure_materialized()

        slices_list = self.properties['slices_list']
        block_indices = self.properties['block_indices']
        arr = self.arr
        if rformat == ReprFormat.NUMPY_ARRAY:
            blocks = [arr]
        else:
            assert isinstance(arr, da.Array)
            blocks = list(map(functools.partial(operator.getitem, arr), slices_list))
        nblocks = len(blocks)
        assert len(slices_list) == nblocks and len(block_indices) == nblocks, \
            (f'Expected equal length of block, indices and slices, got len(blocks)={nblocks} '
             f'len(block_indices)={len(block_indices)} len(slices_list)={len(slices_list)}')

        self.arr = {
            block_indices[i]: (blocks[i], slices_list[i]) for i in range(len(blocks))
        }

        self.properties['repr_format'] = ReprFormat.DICT_BLOCK_INDEX_SLICES

    def reduce(self, force_numpy: bool = False) -> npt.NDArray | da.Array:
        """Concatenate all blocks on the first axis

        Args:
            force_numpy: If True, the result will be forced from dask to a numpy array, if not already; useful for
                outputting to analysis that requires numpy input

        Returns:
            The concatenated result, is Numpy if previous array is Numpy, or if force_numpy is True
        """
        other = NDBlock(self)
        other.ensure_materialized(force_numpy=force_numpy)

        if other.properties['repr_format'] == ReprFormat.NUMPY_ARRAY:
            return np.copy(other.arr)
        else:
            other.to_dict_block_index_slices()

            if other.is_numpy():
                blocks = [block for _, (block, _) in other.arr.items()]
                assert len(blocks) > 0, 'Need at least one row for NDBlock to be reduced'
                assert isinstance(blocks[0], np.ndarray), f'Expected ndarray, got type(blocks[0])={type(blocks[0])}'
                return np.concatenate(blocks, axis=0)
            else:
                shape = None
                for block_index, (block, slices) in other.arr.items():
                    if not isinstance(block, np.ndarray):
                        block = block.compute()
                    shape = (np.nan,) + block.shape[1:]
                    break
                blocks = [da.from_delayed(block,
                                          shape=shape,
                                          dtype=other.get_dtype())
                          for block_index, (block, slices) in other.arr.items()]
                assert len(blocks) > 0, 'Need at least one row for NDBlock to be reduced'

                reduced = da.concatenate(blocks, axis=0)
                if force_numpy:
                    return reduced.compute()
                else:
                    return reduced

    def sum(self, axis: Sequence | None = None, keepdims: bool = False):
        """sum over axes for each block"""
        new_ndblock = NDBlock(self)
        new_ndblock.ensure_materialized()
        if self.properties['repr_format'] == ReprFormat.DICT_BLOCK_INDEX_SLICES:
            for block_index, (block, slices) in self.arr.items():
                new_ndblock.arr[block_index] = (block.sum(axis=axis, keepdims=keepdims), slices)
            return new_ndblock
        else:
            new_ndblock.arr = self.arr.sum(axis=axis)

    @staticmethod
    def map_ndblocks(
            inputs: Sequence[NDBlock],
            fn: Callable,
            out_dtype: np.dtype,
            use_input_index_as_arrloc: int = 0,
            new_slices: list = None,
            fn_args: dict = None
    ) -> NDBlock:
        """Similar to da.map_blocks, but works with NDBlock.

        Different from dask array's map_blocks: For each input i, the block_info[i] provided to the mapping function
        will contain only two keys, 'chunk-location' and 'array-location'.

        Args:
            inputs: A list of inputs to be mapped, either all dask or all numpy; All inputs must have the same number
                of blocks, block indices, and over the same slices as well if inputs are dask images
            fn: fn(\*block, block_info) maps input blocks to an output block
            out_dtype: Output block type (provide a Numpy dtype to this)
            use_input_index_as_arrloc: output slices_list will be the same as this input (ignore this for variable
                sized blocks)
            new_slices: will be used to replace the slices attribute if specified
            fn_args: extra arguments to be passed to the mapping function

        Returns:
            Result mapped, of format ReprFormat.DICT_BLOCK_INDEX_SLICES
        """
        if fn_args is None:
            fn_args = dict()

        inputs = list(inputs)
        assert len(inputs) > 0, 'Must have at least one input!'

        is_numpy = inputs[0].get_repr_format() == ReprFormat.NUMPY_ARRAY
        for i in range(1, len(inputs)):
            assert is_numpy == (inputs[i].get_repr_format() == ReprFormat.NUMPY_ARRAY), \
                ('All inputs must either be all dask or all Numpy, expected '
                 f'is_numpy={is_numpy} but found wrong typed input at '
                 f'location {i}')

        if is_numpy:
            block_info = []
            for i in range(len(inputs)):
                ndblock = inputs[i]
                arr = ndblock.as_numpy()
                block_info.append({
                    'chunk-location': (0,) * arr.ndim,
                    'array-location': list((0, s) for s in arr.shape)
                })
            return NDBlock(fn(*inputs, block_info, **fn_args))

        # we can assume the inputs are all dask, now turn them all into block iterator
        block_iterators = []
        for ndblock in inputs:
            new_block = NDBlock(ndblock)
            new_block.ensure_materialized()
            if ndblock.get_repr_format() != ReprFormat.DICT_BLOCK_INDEX_SLICES:
                new_block.to_dict_block_index_slices()
            block_iterators.append(new_block)

        # write in this form mostly for debugging convenience, equivalently:
        # delayed_fn = dask.delayed(fn)
        @dask.delayed
        def delayed_fn(*blocks, block_info, **kwargs):
            return fn(*blocks, block_info, **kwargs)

        result = {}
        for block_index in block_iterators[use_input_index_as_arrloc].arr.keys():
            blocks = []
            block_info = []
            for i in range(len(block_iterators)):
                block, slices = block_iterators[i].arr[block_index]
                blocks.append(block)
                block_info.append({
                    'chunk-location': block_index,
                    'array-location': slices,
                })
            result[block_index] = (delayed_fn(*blocks, block_info=block_info, **fn_args),
                                   block_info[use_input_index_as_arrloc]['array-location'])

        ndim = block_iterators[0].get_ndim()
        numblocks = block_iterators[0].get_numblocks()
        result_ndblock = NDBlock(None)
        result_ndblock.arr = result
        if new_slices is None:
            slices_list = block_iterators[use_input_index_as_arrloc].properties['slices_list']
        else:
            slices_list = new_slices
        result_ndblock.properties = dict(
            repr_format=ReprFormat.DICT_BLOCK_INDEX_SLICES,
            ndim=ndim,
            numblocks=numblocks,
            block_indices=[k for k in block_iterators[use_input_index_as_arrloc].arr.keys()],
            slices_list=slices_list,
            is_numpy=False,
            dtype=out_dtype
        )
        NDBlock.properties_consistency_check(result_ndblock.properties)
        return result_ndblock

    def select_columns(self, cols: slice | Sequence[int] | int) -> NDBlock:
        """Performs columns selection on a 2d array"""
        ndblock = NDBlock(self)
        ndblock.ensure_materialized()
        if self.properties['repr_format'] == ReprFormat.DICT_BLOCK_INDEX_SLICES:
            results = {
                block_index: (block[:, cols], slices) for block_index, (block, slices) in ndblock.arr.items()
            }
            ndblock.arr = results
        else:
            ndblock.arr = ndblock.arr[:, cols]
        return ndblock

    def run_delayed_discard_results(self):
        assert (self.properties['repr_format'] == ReprFormat.DICT_BLOCK_INDEX_SLICES and
                not self.properties['is_numpy'])
        for value in self.arr.values():
            value[0]()



from typing import Sequence

import dask.array as da
import numpy as np

from cvpl_tools.im.fs import CachePointer

# ----------------------------------- Part 1: rescale, pad and crop -------------------------------------------
"""
The main use case this section of code is: The user tries to resize and fit a smaller 
or larger mask array to another image array by matching the shape of the two images. This consists of two 
steps:
1. Upsample or downsample the mask to match image
2. When the upsample or downsample factor is not a whole integer, add or crop off some of the border pixels
"""


def pad_crop(array: da.Array,
             change_width: Sequence | int,
             mode='constant',
             **kwargs):
    """Similar to dask.array.pad, but allows negative change_width"""
    NDIM = array.ndim
    pad_tups = np.zeros((NDIM, 2), dtype=np.int64)
    if isinstance(change_width, int):
        pad_tups[:] = change_width
    else:
        change_width = tuple(change_width)
        if isinstance(change_width[0], int):
            # shallow change_width, wrap it in a tuple
            change_width = (change_width,)
        for i in range(array.ndim):
            if len(change_width) > i:
                pad_tup = change_width[i]
            else:
                pad_tup = change_width[-1]
            pad_tups[i, 0] = pad_tup[0]
            pad_tups[i, 1] = pad_tup[1]

    to_pad = pad_tups * (pad_tups > 0)
    to_crop = -pad_tups * (pad_tups < 0)

    padded = da.pad(array, to_pad.tolist(), mode, **kwargs)
    padded_shape = padded.shape
    crop_slices = tuple(slice(to_crop[i, 0], padded_shape[i] - to_crop[i, 1]) for i in range(NDIM))

    return padded[crop_slices]


def approx_upsample_scale(src_shape: tuple[int, ...],
                          tgt_shape: tuple[int, ...]
                          ) -> tuple[list[int, ...], list[int, ...]]:
    scale = []
    residue = []

    for i in range(len(src_shape)):
        ss, ts = src_shape[i], tgt_shape[i]
        s = max(round(ts / ss), 1)
        r = ts - ss * s
        scale.append(s)
        residue.append(r)
    return scale, residue


def approx_downpsample_scale(src_shape: tuple[int, ...],
                             tgt_shape: tuple[int, ...]
                             ) -> tuple[list[int, ...], list[int, ...]]:
    """The residue is calculated assuming scale is supplied as factor to the any_to_any.DownsamplingByIntFactor class"""
    scale = []
    residue = []

    for i in range(len(src_shape)):
        ss, ts = src_shape[i], tgt_shape[i]
        s = max(round(ss / ts), 1)
        r = ts - ((ss - 1) // s + 1)
        scale.append(s)
        residue.append(r)
    return scale, residue


def upsample_pad_crop_fit(src_arr: da.Array, tgt_arr: da.Array, cptr: CachePointer,
                          viewer_args: dict, mode='constant', **kwargs):
    import cvpl_tools.im.process.any_to_any as sp_any_to_any
    src_shape = src_arr.shape
    tgt_shape = tgt_arr.shape
    scale, residue = approx_upsample_scale(src_shape, tgt_shape)
    upsampler = sp_any_to_any.UpsamplingByIntFactor(factor=tuple(scale), order=0)
    arr = upsampler.forward(src_arr, cptr=cptr, viewer_args=viewer_args)

    change_width = tuple((0, -r) for r in residue)
    arr = pad_crop(arr, change_width=change_width, mode=mode, **kwargs)
    return arr


if __name__ == '__main__':
    test_arr = da.zeros((4, 4), dtype=np.float32)
    test_arr[1, 1] = 1
    test_arr2 = pad_crop(test_arr, change_width=-1, mode='constant')
    from numpy.testing import assert_almost_equal

    assert_almost_equal(test_arr2,
                        np.array(
                            ((1., 0.),
                             (0., 0.)),
                            dtype=np.float32
                        ))



import zmq
import logging
import socket
from contextlib import contextmanager, suppress
import os
from threading import Thread, Lock
import shutil

from partd.dict import Dict
from partd.buffer import Buffer
import sqlite3
from partd.core import Interface
import locket
from dask.distributed import print as dprint


class SQLiteKVStore:
    def __init__(self, db_path: str):
        self.is_exists = os.path.exists(db_path)  # may not be accurate
        self.conn = sqlite3.connect(db_path)
        self.cursor = self.conn.cursor()
        assert self.cursor is not None
        self.write_row_stmt = None
        self.read_row_stmt = None
        self.init_db()

    def init_db(self):
        if not self.is_exists:
            self.cursor.execute('''
            CREATE TABLE IF NOT EXISTS kv_store (
                id TEXT PRIMARY KEY,
                value TEXT
            )
            ''')
        self.write_row_stmt = """
        INSERT INTO kv_store (id, value) VALUES (?, ?)
        ON CONFLICT(id) DO UPDATE SET value=excluded.value;
        """
        self.read_row_stmt = """
        SELECT value FROM kv_store WHERE id = ?
        """

    def write_rows(self, tups):
        self.cursor.executemany(self.write_row_stmt, tups)

    def write_row(self, tup):
        self.cursor.execute(self.write_row_stmt, tup)

    def read_rows(self, ids):
        return [self.read_row(id) for id in ids]

    def read_row(self, id):
        self.cursor.execute(self.read_row_stmt, (id,))
        result = self.cursor.fetchone()

        result = result[0]
        return result

    def commit(self):
        self.conn.commit()

    def close(self):
        self.conn.close()


# ------------------------------------------Part 2: Partd Class-----------------------------------------------
"""
Copyright (c) 2015, Continuum Analytics, Inc. and contributors
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of Continuum Analytics nor the names of any contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE.

SQLitePartd class is modified from the partd class below:
reference: https://github.com/dask/partd/blob/main/partd/file.py commit hash: efa78b4
"""


class SQLitePartd(Interface):
    def __init__(self, path: str, create_kv_store=None):
        """path is an directory under which the lock file will be placed"""
        self.path = path
        os.makedirs(path, exist_ok=True)
        if create_kv_store is None:
            def create_kv_store():
                return SQLiteKVStore(f'{path}/sqlite.db')
        self.kv_store = create_kv_store()
        self.create_kv_store = create_kv_store
        self.lock = locket.lock_file(f'{path}/.lock')
        Interface.__init__(self)

    def __getstate__(self):
        return {'path': self.path}

    def __setstate__(self, state):
        Interface.__setstate__(self, state)
        self.__class__.__init__(self, state['path'])

    def append(self, data, lock=True, fsync=False, **kwargs):
        if lock:
            self.lock.acquire()
        try:
            self.kv_store.write_rows(data.items())
            self.kv_store.commit()
        finally:
            if lock:
                self.lock.release()

    def _get(self, keys, lock=True, **kwargs):
        assert isinstance(keys, (list, tuple, set))
        if lock:
            self.lock.acquire()
        try:
            if len(keys) == 1:
                result = [self.kv_store.read_row(keys[0])]
            else:
                result = self.kv_store.read_rows(keys)
        finally:
            if lock:
                self.lock.release()
        return result

    def _iset(self, key, value, lock=True):
        """ Idempotent set """
        if lock:
            self.lock.acquire()
        try:
            self.kv_store.write_row((key, value))
            self.kv_store.commit()
        finally:
            if lock:
                self.lock.release()

    def _delete(self, keys, lock=True):
        if lock:
            self.lock.acquire()
        try:
            raise NotImplementedError()
        finally:
            if lock:
                self.lock.release()

    def drop(self):
        self.kv_store.close()
        if os.path.exists(self.path):
            shutil.rmtree(self.path)
        self._iset_seen.clear()
        os.mkdir(self.path)
        self.kv_store = self.create_kv_store()

    def close(self):
        self.kv_store.close()

    def __exit__(self, *args):
        self.drop()
        os.rmdir(self.path)
        self.kv_store.close()


# --------------------------------------------Part 3: Server------------------------------------------------


tuple_sep = b'-|-'


logger = logging.getLogger(__name__)


@contextmanager
def logerrors():
    try:
        yield
    except Exception as e:
        logger.exception(e)
        raise


class SqliteServer:
    """
    SqliteServer class is modified from the Server/Client class file below:
    https://github.com/dask/partd/blob/main/partd/zmq.py
    """
    def __init__(self, path, nappend, available_memory=None, get_sqlite_partd=None, port_protocol='tcp'):
        """
        Args:
            path:
            available_memory:
            get_sqlite_partd:
            nappend: Number of messages to append, after which point the server closes
            port_protocol: 'tcp' for network or 'ipc' for linux
        """

        self.path = path
        self.available_memory = available_memory
        if get_sqlite_partd is None:
            def get_sqlite_partd():
                return SQLitePartd(path)
        self._get_sqlite_partd = get_sqlite_partd

        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.ROUTER)
        hostname = socket.gethostname()

        if port_protocol == 'tcp':
            port = self.socket.bind_to_random_port('tcp://*')
            self.address = ('tcp://%s:%d' % (hostname, port)).encode()
        elif port_protocol == 'ipc':
            import tempfile
            temp_dir = tempfile.gettempdir()
            ipc_path = os.path.join(temp_dir, f"zmq_ipc_{os.getpid()}")
            self.address = f"ipc://{ipc_path}"
        else:
            raise ValueError(f'port_protocol must be either "tcp" or "ipc", but found {port_protocol}')

        self.status = 'run'
        assert isinstance(nappend, int), f'Expected int, got type(nappend)={type(nappend)}'
        self.nappend = nappend
        self.nappended = 0
        self._socket_lock = Lock()
        self._listen_thread = Thread(target=self.listen)
        self._listen_thread.start()
        logger.debug('Start server at %s', self.address)

    def get_partd(self):
        partd = self._get_sqlite_partd()
        if self.available_memory is not None:
            buffer_partd = Buffer(Dict(), partd, available_memory=self.available_memory)
        else:
            buffer_partd = partd
        return partd, buffer_partd

    def wait_join(self):
        """ Block until all threads close """
        try:
            self._listen_thread.join()
        except AttributeError:
            pass

    def listen(self):
        with logerrors():
            logger.debug('Start listening %s', self.address)
            partd, buffer_partd = self.get_partd()
            while self.status != 'closed':
                if not self.socket.poll(100):
                    continue
                with self._socket_lock:
                    payload = self.socket.recv_multipart()

                address, command, payload = payload[0], payload[1], payload[2:]
                logger.debug('Server receives %s %s', address, command)

                if command == b'close':
                    self.ack(address)
                    break

                elif command == b'append':
                    keys, values = payload[::2], payload[1::2]
                    keys = list(map(deserialize_key, keys))
                    data = dict(zip(keys, values))
                    buffer_partd.append(data, lock=False)
                    logger.debug('Server appends %d keys', len(data))
                    self.ack(address)
                    self.nappended += 1
                    if self.nappend == self.nappended:
                        dprint(f'nappended reached nappend={self.nappend}')
                        break

                elif command == b'iset':
                    key, value = payload
                    key = deserialize_key(key)
                    buffer_partd.iset(key, value, lock=False)
                    self.ack(address)

                elif command == b'get':
                    keys = list(map(deserialize_key, payload))
                    logger.debug('get %s', keys)
                    result = buffer_partd.get(keys)
                    self.send_to_client(address, result)
                    self.ack(address, flow_control=False)

                elif command == b'syn':
                    self.ack(address)

                else:
                    logger.debug("Unknown command: %s", command)
                    raise ValueError("Unknown command: " + str(command))

            if buffer_partd != partd:
                buffer_partd.flush()
            partd.close()

    def send_to_client(self, address, result):
        with logerrors():
            if not isinstance(result, list):
                result = [result]
            with self._socket_lock:
                self.socket.send_multipart([address] + result)

    def ack(self, address, flow_control=True):
        with logerrors():
            logger.debug('Server sends ack')
            self.send_to_client(address, b'ack')

    def close(self):
        logger.debug('Server closes')
        print('server properly closed')
        self.wait_join()
        self.status = 'closed'
        with suppress(zmq.error.ZMQError):
            self.socket.close(1)
        with suppress(zmq.error.ZMQError):
            self.context.destroy(3)


def deserialize_key(text):
    """
    >>> deserialize_key(b'x')
    b'x'
    >>> deserialize_key(b'a-|-b-|-1')
    (b'a', b'b', b'1')
    """
    if tuple_sep in text:
        return tuple(text.split(tuple_sep))
    else:
        return text
